
#include "relic_fp_low.h"
.global fp_add_low		// Modular addition
//.global fp_addd_low             // Double-precision modular addition in [0, p^2]
.global fp_sub_low              // Modular subtraction
.global fp_subd_low		// Double-precision modular subtraction in [0, p^2]	

.global fp_muln_low_asm_basic_256   // School-book multiplication (optimized for FP_PRIME = 256).text
.global fp_muln_low_asm_comba_256   // Coma multiplication (optimized for FP_PRIME = 256)

.text
.arm
	
/*
 * Function: fp_add_low
 * Inputs: r1 = a, r2 = b
 * Output: r0 = c = (a+b) mod p
 */


#define P0 0x00000013
#define P1 0xA7000000
#define P2 0x00000013
#define P3 0x61210000
#define P4 0x00000008
#define P5 0xBA344D80
#define P6 0x40000001
#define P7 0x25236482


fp_add_low:
	push	{r5-r12}
	ldr     r4, [r1]
	ldr     r12, [r2]
	adds	r4, r4, r12
	str	r4, [r0]
	
	ldr     r5, [r1, #4]
	ldr     r12, [r2, #4]
	adcs    r5, r5, r12
	str     r5, [r0, #4]

        ldr     r6, [r1, #8]
	ldr     r12, [r2, #8]
        adcs    r6, r6, r12
        str     r6, [r0, #8]

        ldr     r7, [r1, #12]
        ldr     r12, [r2, #12]
        adcs    r7, r7, r12
        str     r7, [r0, #12]

	ldr     r8, [r1, #16]
        ldr     r12, [r2, #16]
        adcs    r8, r8, r12
        str     r8, [r0, #16]

        ldr     r9, [r1, #20]
        ldr     r12, [r2, #20]
        adcs    r9, r9, r12
        str     r9, [r0, #20]

        ldr     r10, [r1, #24]
        ldr     r12, [r2, #24]
        adcs    r10, r10, r12
        str     r10, [r0, #24]

        ldr     r11, [r1, #28]
        ldr     r12, [r2, #28]
        adcs     r11, r11, r12
        str     r11, [r0, #28]

        //ldr     r12, [r3]
	subs    r4, r4, #P0
        //ldr     r12, [r3, #4]
        sbcs    r5, r5, #P1
        //ldr     r12, [r3, #8]
        sbcs    r6, r6, #P2
        ldr     r12, [r3, #12]
        sbcs    r7, r7, r12
        //ldr     r12, [r3, #16]
        sbcs    r8, r8, #P4
        ldr     r12, [r3, #20]
        sbcs    r9, r9, r12
        //ldr     r12, [r3, #24]
        sbcs    r10, r10, #P6
        ldr     r12, [r3, #28]
        sbcs    r11, r11, r12

	bcc	exit1
	stmia	r0, {r4-r11}
exit1:	
	pop     {r5-r12}
	mov	pc, lr

/*
fp_addd_low:
        push    {r5-r12}
        ldr     r4, [r1]
        ldr     r12, [r2]
        adds    r4, r4, r12
        str     r4, [r0]

        ldr     r5, [r1, #4]
        ldr     r12, [r2, #4]
        adcs    r5, r5, r12
        str     r5, [r0, #4]

        ldr     r6, [r1, #8]
        ldr     r12, [r2, #8]
        adcs    r6, r6, r12
        str     r6, [r0, #8]

        ldr     r7, [r1, #12]
        ldr     r12, [r2, #12]
        adcs    r7, r7, r12
        str     r7, [r0, #12]

        ldr     r8, [r1, #16]
        ldr     r12, [r2, #16]
        adcs    r8, r8, r12
        str     r8, [r0, #16]

        ldr     r9, [r1, #20]
        ldr     r12, [r2, #20]
        adcs    r9, r9, r12
        str     r9, [r0, #20]

        ldr     r10, [r1, #24]
        ldr     r12, [r2, #24]
        adcs    r10, r10, r12
        str     r10, [r0, #24]

        ldr     r11, [r1, #28]
        ldr     r12, [r2, #28]
        adcs    r11, r11, r12
        str     r11, [r0, #28]

        ldr     r4, [r1, #32]
        ldr     r12, [r2, #32]
        adcs    r4, r4, r12
        str     r4, [r0, #32]

        ldr     r5, [r1, #36]
        ldr     r12, [r2, #36]
        adcs    r5, r5, r12
        str     r5, [r0, #36]

        ldr     r6, [r1, #40]
        ldr     r12, [r2, #40]
        adcs    r6, r6, r12
        str     r6, [r0, #40]

        ldr     r7, [r1, #44]
        ldr     r12, [r2, #44]
        adcs    r7, r7, r12
        str     r7, [r0, #44]

        ldr     r8, [r1, #48]
        ldr     r12, [r2, #48]
        adcs    r8, r8, r12
        str     r8, [r0, #48]

        ldr     r9, [r1, #52]
        ldr     r12, [r2, #52]
        adcs    r9, r9, r12
        str     r9, [r0, #52]

        ldr     r10, [r1, #56]
        ldr     r12, [r2, #56]
        adcs    r10, r10, r12
        str     r10, [r0, #56]

        ldr     r11, [r1, #60]
        ldr     r12, [r2, #60]
        adcs    r11, r11, r12
        str     r11, [r0, #60]

        ldr     r4, [r0]
	ldr     r12, [r3]
        subs    r4, r4, r12

        ldr     r5, [r0, #4]
        ldr     r12, [r3, #4]
        sbcs    r5, r5, r12

        ldr     r6, [r0, #8]
        ldr     r12, [r3, #8]
        sbcs    r6, r6, r12

        ldr     r7, [r0, #12]
        ldr     r12, [r3, #12]
        sbcs    r7, r7, r12

        ldr     r8, [r0, #16]
        ldr     r12, [r3, #16]
        sbcs    r8, r8, r12

        ldr     r9, [r0, #20]
        ldr     r12, [r3, #20]
        sbcs    r9, r9, r12

        ldr     r10, [r0, #24]
        ldr     r12, [r3, #24]
        sbcs    r10, r10, r12

        ldr     r11, [r0, #28]
        ldr     r12, [r3, #28]
        sbcs    r11, r11, r12

        ldr     r4, [r0, #32]
        ldr     r12, [r3, #32]
        sbcs    r4, r4, r12

        ldr     r5, [r0, #36]
        ldr     r12, [r3, #36]
        sbcs    r5, r5, r12

        ldr     r6, [r0, #40]
        ldr     r12, [r3, #40]
        sbcs    r6, r6, r12

        ldr     r7, [r0, #44]
        ldr     r12, [r3, #44]
        sbcs    r7, r7, r12

        ldr     r8, [r0, #48]
        ldr     r12, [r3, #48]
        sbcs    r8, r8, r12

        ldr     r9, [r0, #52]
        ldr     r12, [r3, #52]
        sbcs    r9, r9, r12

        ldr     r10, [r0, #56]
        ldr     r12, [r3, #56]
        sbcs    r10, r10, r12

        ldr     r11, [r0, #60]
        ldr     r12, [r3, #60]
        sbcs    r11, r11, r12

	bcc     exit2

        ldr     r4, [r0]
        ldr     r12, [r3]
        subs    r4, r4, r12

        ldr     r5, [r0, #4]
        ldr     r12, [r3, #4]
        sbcs    r5, r5, r12

        ldr     r6, [r0, #8]
        ldr     r12, [r3, #8]
        sbcs    r6, r6, r12

        ldr     r7, [r0, #12]
        ldr     r12, [r3, #12]
        sbcs    r7, r7, r12

        ldr     r8, [r0, #16]
        ldr     r12, [r3, #16]
        sbcs    r8, r8, r12

        ldr     r9, [r0, #20]
        ldr     r12, [r3, #20]
        sbcs    r9, r9, r12

        ldr     r10, [r0, #24]
        ldr     r12, [r3, #24]
        sbcs    r10, r10, r12

        ldr     r11, [r0, #28]
        ldr     r12, [r3, #28]
        sbcs    r11, r11, r12
	
	stmia   r0!, {r4-r11}

        ldr     r4, [r0, #32]
        ldr     r12, [r3, #32]
        sbcs    r4, r4, r12

        ldr     r5, [r0, #36]
        ldr     r12, [r3, #36]
        sbcs    r5, r5, r12

        ldr     r6, [r0, #40]
        ldr     r12, [r3, #40]
        sbcs    r6, r6, r12

        ldr     r7, [r0, #44]
        ldr     r12, [r3, #44]
        sbcs    r7, r7, r12

        ldr     r8, [r0, #48]
        ldr     r12, [r3, #48]
        sbcs    r8, r8, r12

        ldr     r9, [r0, #52]
        ldr     r12, [r3, #52]
        sbcs    r9, r9, r12

        ldr     r10, [r0, #56]
        ldr     r12, [r3, #56]
        sbcs    r10, r10, r12

        ldr     r11, [r0, #60]
        ldr     r12, [r3, #60]
        sbcs    r11, r11, r12

	//add	r0, r0, #0x32
        stmia   r0, {r4-r11}	
exit2:
        pop     {r5-r12}
        mov     pc, lr
*/	

/*
 * Function: fp_sub_low
 * Inputs: r1 = a, r2 = b
 * Output: r0 = c = (a-b) mod p
 */

fp_sub_low:
        push    {r5-r12}
        ldr     r4, [r1]
        ldr     r12, [r2]
        subs    r4, r4, r12
        str     r4, [r0]

        ldr     r5, [r1, #4]
        ldr     r12, [r2, #4]
        sbcs    r5, r5, r12
        str     r5, [r0, #4]

        ldr     r6, [r1, #8]
        ldr     r12, [r2, #8]
        sbcs    r6, r6, r12
        str     r6, [r0, #8]

        ldr     r7, [r1, #12]
        ldr     r12, [r2, #12]
        sbcs    r7, r7, r12
        str     r7, [r0, #12]

        ldr     r8, [r1, #16]
        ldr     r12, [r2, #16]
        sbcs    r8, r8, r12
        str     r8, [r0, #16]

        ldr     r9, [r1, #20]
        ldr     r12, [r2, #20]
        sbcs    r9, r9, r12
        str     r9, [r0, #20]

        ldr     r10, [r1, #24]
        ldr     r12, [r2, #24]
        sbcs    r10, r10, r12
        str     r10, [r0, #24]

	ldr     r11, [r1, #28]
        ldr     r12, [r2, #28]
        sbcs    r11, r11, r12
        str     r11, [r0, #28]

        bcs     exit3

        //ldr     r12, [r3]
        //ldr     r4, [r0]
        adds    r12, r4, #P0
        str     r12, [r0]

        //ldr     r12, [r3, #4]
        //ldr     r4, [r0, #4]
        adcs    r12, r5, #P1
        str     r12, [r0, #4]

        //ldr     r12, [r3, #8]
        //ldr     r4, [r0, #8]
        adcs    r12, r6, #P2
        str     r12, [r0, #8]

        ldr     r12, [r3, #12]
        //ldr     r4, [r0, #12]
        adcs    r12, r12, r7
        str     r12, [r0, #12]

        //ldr     r12, [r3, #16]
        //ldr     r4, [r0, #16]
        adcs    r12, r8, #P4
        str     r12, [r0, #16]

        ldr     r12, [r3, #20]
        //ldr     r4, [r0, #20]
        adcs    r12, r12, r9
        str     r12, [r0, #20]

        //ldr     r12, [r3, #24]
        //ldr     r4, [r0, #24]
        adcs    r12, r10, #P6
        str     r12, [r0, #24]

        ldr     r12, [r3, #28]
        //ldr     r4, [r0, #28]
        adcs    r12, r12, r11
        str     r12, [r0, #28]
exit3:
        pop     {r5-r12}
        mov     pc, lr

	
fp_subd_low:
        push    {r5-r12}
        ldr     r4, [r1]
        ldr     r12, [r2]
        subs    r4, r4, r12
        str     r4, [r0]

        ldr     r5, [r1, #4]
        ldr     r12, [r2, #4]
        sbcs    r5, r5, r12
        str     r5, [r0, #4]

        ldr     r6, [r1, #8]
        ldr     r12, [r2, #8]
        sbcs    r6, r6, r12
        str     r6, [r0, #8]

        ldr     r7, [r1, #12]
        ldr     r12, [r2, #12]
        sbcs    r7, r7, r12
        str     r7, [r0, #12]

        ldr     r8, [r1, #16]
        ldr     r12, [r2, #16]
        sbcs    r8, r8, r12
        str     r8, [r0, #16]

        ldr     r9, [r1, #20]
        ldr     r12, [r2, #20]
        sbcs    r9, r9, r12
        str     r9, [r0, #20]

        ldr     r10, [r1, #24]
        ldr     r12, [r2, #24]
        sbcs    r10, r10, r12
        str     r10, [r0, #24]

        ldr     r11, [r1, #28]
        ldr     r12, [r2, #28]
        sbcs    r11, r11, r12
        str     r11, [r0, #28]
	
        ldr     r4, [r1, #32]
        ldr     r12, [r2, #32]
        sbcs    r4, r4, r12
        str     r4, [r0, #32]

        ldr     r5, [r1, #36]
        ldr     r12, [r2, #36]
        sbcs    r5, r5, r12
        str     r5, [r0, #36]

        ldr     r6, [r1, #40]
        ldr     r12, [r2, #40]
        sbcs    r6, r6, r12
        str     r6, [r0, #40]

        ldr     r7, [r1, #44]
        ldr     r12, [r2, #44]
        sbcs    r7, r7, r12
        str     r7, [r0, #44]

	ldr     r8, [r1, #48]
        ldr     r12, [r2, #48]
        sbcs    r8, r8, r12
        str     r8, [r0, #48]

        ldr     r9, [r1, #52]
        ldr     r12, [r2, #52]
        sbcs    r9, r9, r12
        str     r9, [r0, #52]

        ldr     r10, [r1, #56]
        ldr     r12, [r2, #56]
        sbcs    r10, r10, r12
        str     r10, [r0, #56]

        ldr     r11, [r1, #60]
        ldr     r12, [r2, #60]
        sbcs    r11, r11, r12
        str     r11, [r0, #60]

	bcs 	exit4

	ldr     r12, [r3]
	ldr     r11, [r0]
        adds    r12, r11, r12
	str	r12, [r0]
	
        ldr     r12, [r3, #4]
        ldr     r11, [r0, #4]
        adcs    r12, r12, r11
        str     r12, [r0, #4]
	
        ldr     r12, [r3, #8]
        ldr     r11, [r0, #8]
        adcs    r12, r12, r11
        str     r12, [r0, #8]

        ldr     r12, [r3, #12]
        ldr     r11, [r0, #12]
        adcs    r12, r12, r11
        str     r12, [r0, #12]

        ldr     r12, [r3, #16]
        ldr     r11, [r0, #16]
        adcs    r12, r12, r11
        str     r12, [r0, #16]

        ldr     r12, [r3, #20]
        ldr     r11, [r0, #20]
        adcs    r12, r12, r11
        str     r12, [r0, #20]

        ldr     r12, [r3, #24]
        ldr     r11, [r0, #24]
        adcs    r12, r12, r11
        str     r12, [r0, #24]

        ldr     r12, [r3, #28]
        ldr     r11, [r0, #28]
        adcs    r12, r12, r11
        str     r12, [r0, #28]

        ldr     r12, [r3, #32]
        //ldr     r4, [r0, #32]
        adcs    r12, r12, r4
        str     r12, [r0, #32]

        ldr     r12, [r3, #36]
        //ldr     r4, [r0, #36]
        adcs    r12, r12, r5
        str     r12, [r0, #36]

	ldr     r12, [r3, #40]
        //ldr     r4, [r0, #40]
        adcs    r12, r12, r6
        str     r12, [r0, #40]

        ldr     r12, [r3, #44]
        //ldr     r4, [r0, #44]
        adcs    r12, r12, r7
        str     r12, [r0, #44]

        ldr     r12, [r3, #48]
        //ldr     r4, [r0, #48]
        adcs    r12, r12, r8
        str     r12, [r0, #48]

        ldr     r12, [r3, #52]
        //ldr     r4, [r0, #52]
        adcs    r12, r12, r9
        str     r12, [r0, #52]

	ldr     r12, [r3, #56]
        //ldr     r4, [r0, #56]
        adcs    r12, r12, r10
        str     r12, [r0, #56]

        ldr     r12, [r3, #60]
        ldr     r11, [r0, #60]
        adcs    r12, r12, r11
        str     r12, [r0, #60]	
exit4:  
        pop     {r5-r12}
	mov	pc, lr

	
	
	
	
	
	
/*
 * Function: fp_muln_low_asm_basic_256
 * Inputs: r1 = a, r2 = b
 * Output: r0 = c = a*b
 */

fp_muln_low_asm_basic_256:
	push	{r0-r12}
	push	{r14}
	
	ldr     r3, [r1]
	bl		compute_partial_product
	stmia	r0, {r4-r12}
	add		r0, r0, #4

	ldr     r3, [r1, #4]
	bl		compute_partial_product
	bl		accumulate
	stmia	r0, {r4-r12}
	add		r0, r0, #4

	ldr     r3, [r1, #8]
	bl		compute_partial_product
	bl		accumulate
	stmia	r0, {r4-r12}
	add		r0, r0, #4

	ldr     r3, [r1, #12]
	bl		compute_partial_product
	bl		accumulate
	stmia	r0, {r4-r12}
	add		r0, r0, #4

	ldr     r3, [r1, #16]
	bl		compute_partial_product
	bl		accumulate
	stmia	r0, {r4-r12}
	add		r0, r0, #4

	ldr     r3, [r1, #20]
	bl		compute_partial_product
	bl		accumulate
	stmia	r0, {r4-r12}
	add		r0, r0, #4

	ldr     r3, [r1, #24]
	bl		compute_partial_product
	bl		accumulate
	stmia	r0, {r4-r12}
	add		r0, r0, #4

	ldr     r3, [r1, #28]
	bl		compute_partial_product
	bl		accumulate
	stmia	r0, {r4-r12}

	pop		{r14}
	pop		{r0-r12}
	mov		pc, lr


clear_working_regs:
//	eor		r5, r5, r5
	eor		r6, r6, r6
	eor		r7, r7, r7
	eor		r8, r8, r8
	eor		r9, r9, r9
	eor		r10, r10, r10
	eor		r11, r11, r11
	eor		r12, r12, r12
	mov		pc, lr
	
compute_partial_product:
	push 	{r14}

	eor		r6, r6, r6
	eor		r7, r7, r7
	eor		r8, r8, r8
	eor		r9, r9, r9
	eor		r10, r10, r10
	eor		r11, r11, r11
	eor		r12, r12, r12
	
	ldr     r14, [r2]
	umull	r4, r5, r3, r14
	
	ldr		r14, [r2, #4]
	umlals	r5, r6, r3, r14

 	ldr		r14, [r2, #8]
	umlals	r6, r7, r3, r14
	
	ldr		r14, [r2, #12]
	umlals	r7, r8, r3, r14
	
	ldr		r14, [r2, #16]
	umlals	r8, r9, r3, r14
	
	ldr		r14, [r2, #20]
	umlals	r9, r10, r3, r14
	
	ldr		r14, [r2, #24]
	umlals	r10, r11, r3, r14
	
	ldr		r14, [r2, #28]
	umlals	r11, r12, r3, r14

	pop		{r14}
	mov		pc, lr
	

accumulate:
	ldr		r3, [r0]
	adds	r4, r4, r3
	ldr		r3, [r0, #4]
	adcs	r5, r5, r3
	ldr		r3, [r0, #8]
	adcs	r6, r6, r3
	ldr		r3, [r0, #12]
	adcs	r7, r7, r3
	ldr		r3, [r0, #16]
	adcs	r8, r8, r3
	ldr		r3, [r0, #20]
	adcs	r9, r9, r3
	ldr		r3, [r0, #24]
	adcs	r10, r10, r3
	ldr		r3, [r0, #28]
	adcs	r11, r11, r3
	adcs	r12, r12, #0


	mov		pc, lr



	


/*
 * Function: fp_muln_low_asm_comba_256
 * Inputs: r1 = a, r2 = b
 * Output: r0 = c = a*b
 */

fp_muln_low_asm_comba_256:
	
	push	{r0-r12} 
	push	{r14}
	
	pld		[r1, #32]
	pld		[r2, #32]
	

	/* col 0 */
	ldr     r12, [r1]
	ldr		r14, [r2]
		mov		r5, #0
	umull	r3, r4, r12, r14
		mov		r11, #0
		mov		r6, #0
	
	/* col 1 */
	ldr		r14, [r2, #4]
		mov		r10, #0
	umlal	r4, r5, r12, r14
		mov		r7, #0
	
	ldr     r12, [r1, #4]
	ldr		r14, [r2]
		mov		r9, #0
	umlal	r4, r11, r12, r14

		mov		r8, #0
		ldr     r12, [r1, #8]
	
	adds	r5, r5, r11
	adc		r6, r6, #0
	
	/* col 2 */
	umlal	r5, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #4]
		ldr		r14, [r2, #4]
	
	adds	r6, r6, r10
	adc		r7, r7, #0

	umlal	r5, r9, r12, r14

		mov		r10, #0
		ldr     r12, [r1]
		ldr		r14, [r2, #8]

	adds	r6, r6, r9
	adc		r7, r7, #0
	
	umlal	r5, r8, r12, r14

		mov		r9, #0
		ldr		r14, [r2, #12]

	adds	r6, r6, r8
	adcs	r7, r7, #0


	/* col 3 */
	umlal	r6, r11, r12, r14

		mov		r8, #0
		ldr     r12, [r1, #4]
		ldr		r14, [r2, #8]

	adds	r7, r7, r11
	adc		r8, r8, #0

	umlal	r6, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #8]
		ldr		r14, [r2, #4]

	adds	r7, r7, r10
	adc		r8, r8, #0
	
	umlal	r6, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #12]
		ldr		r14, [r2]

	adds	r7, r7, r11
	adc		r8, r8, #0

	umlal	r6, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #16]

			stmia	r0!, {r3-r6}

	adds	r7, r7, r10
	adc		r8, r8, #0

			mov		r3, r7

			
			
	
	/* col 4 */
	umlal	r3, r11, r12, r14
		
			mov		r4, r8
			mov		r5, #0
		ldr     r12, [r1, #12]
		ldr		r14, [r2, #4]
		mov		r10, #0

	adds	r4, r4, r11
	adc		r5, r5, #0

	umlal	r3, r10, r12, r14
		
			mov		r6, #0
		mov		r11, #0
		ldr     r12, [r1, #8]
		ldr		r14, [r2, #8]
	
	adds	r4, r4, r10
	adc		r5, r5, #0

	umlal	r3, r11, r12, r14

			mov		r7, #0
		mov		r10, #0
		ldr     r12, [r1, #4]
		ldr		r14, [r2, #12]

	adds	r4, r4, r11
	adc		r5, r5, #0

	umlal	r3, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1]
		ldr		r14, [r2, #16]
		
	adds	r4, r4, r10
	adc		r5, r5, #0

	umlal	r3, r11, r12, r14
		
			mov		r8, #0
		mov		r10, #0
		ldr		r14, [r2, #20]
		
	adds	r4, r4, r11
	adc		r5, r5, #0


	/* col 5 */
	umlal	r4, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #4]
		ldr		r14, [r2, #16]

	adds	r5, r5, r10
	adc		r6, r6, #0

	umlal	r4, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #8]
		ldr		r14, [r2, #12]

	adds	r5, r5, r11
	adc		r6, r6, #0
	
	umlal	r4, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #12]
		ldr		r14, [r2, #8]

	adds	r5, r5, r10
	adc		r6, r6, #0
	
	umlal	r4, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #16]
		ldr		r14, [r2, #4]

	adds	r5, r5, r11
	adc		r6, r6, #0

	umlal	r4, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #20]
		ldr		r14, [r2]

	adds	r5, r5, r10
	adc		r6, r6, #0

	
	umlal	r4, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #24]

	adds	r5, r5, r11
	adc		r6, r6, #0

	
	/* col 6 */
	umlal	r5, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #20]
		ldr		r14, [r2, #4]

	adds	r6, r6, r10
	adc		r7, r7, #0

	umlal	r5, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #16]
		ldr		r14, [r2, #8]

	adds	r6, r6, r11
	adc		r7, r7, #0

	umlal	r5, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #12]
		ldr		r14, [r2, #12]

	adds	r6, r6, r10
	adc		r7, r7, #0

	umlal	r5, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #8]
		ldr		r14, [r2, #16]

	adds	r6, r6, r11
	adc		r7, r7, #0

	umlal	r5, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #4]
		ldr		r14, [r2, #20]

	adds	r6, r6, r10
	adc		r7, r7, #0
	
	umlal	r5, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1]
		ldr		r14, [r2, #24]

	adds	r6, r6, r11
	adc		r7, r7, #0

	umlal	r5, r10, r12, r14

		mov		r11, #0
		ldr		r14, [r2, #28]

	adds	r6, r6, r10
	adc		r7, r7, #0


	/* col 7 */
	umlal	r6, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #4]
		ldr		r14, [r2, #24]

	adds	r7, r7, r11
	adc		r8, r8, #0

	umlal	r6, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #8]
		ldr		r14, [r2, #20]

	adds	r7, r7, r10
	adc		r8, r8, #0

	umlal	r6, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #12]
		ldr		r14, [r2, #16]

	adds	r7, r7, r11
	adc		r8, r8, #0

	umlal	r6, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #16]
		ldr		r14, [r2, #12]

	adds	r7, r7, r10
	adc		r8, r8, #0

	umlal	r6, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #20]
		ldr		r14, [r2, #8]

	adds	r7, r7, r11
	adc		r8, r8, #0

	umlal	r6, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #24]
		ldr		r14, [r2, #4]
	
	adds	r7, r7, r10
	adc		r8, r8, #0
	
	umlal	r6, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #28]
		ldr		r14, [r2]

	adds	r7, r7, r11
	adc		r8, r8, #0
	
	umlal	r6, r10, r12, r14
	
		mov		r11, #0
		ldr		r14, [r2, #4]
			
			stmia	r0!, {r3-r6}
	
	adds	r7, r7, r10
	adc		r8, r8, #0

			mov		r3, r7
	

	/* col 8 */
	umlal	r3, r11, r12, r14
	
			mov		r4, r8
			mov		r5, #0
		mov		r10, #0
		ldr     r12, [r1, #24]
		ldr		r14, [r2, #8]

	adds	r4, r4, r11
	adc		r5, r5, #0

	umlal	r3, r10, r12, r14

			mov		r6, #0
		mov		r11, #0
		ldr     r12, [r1, #20]
		ldr		r14, [r2, #12]

	adds	r4, r4, r10
	adc		r5, r5, #0

	umlal	r3, r11, r12, r14

			mov		r7, #0
		mov		r10, #0
		ldr     r12, [r1, #16]
		ldr		r14, [r2, #16]

	adds	r4, r4, r11
	adc		r5, r5, #0

	umlal	r3, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #12]
		ldr		r14, [r2, #20]

	adds	r4, r4, r10
	adc		r5, r5, #0

	umlal	r3, r11, r12, r14

		mov		r10, #0
		ldr     r12, [r1, #8]
		ldr		r14, [r2, #24]

	adds	r4, r4, r11
	adc		r5, r5, #0

	umlal	r3, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #4]
		ldr		r14, [r2, #28]

	adds	r4, r4, r10
	adc		r5, r5, #0

	umlal	r3, r11, r12, r14

			mov		r8, #0
		mov		r10, #0
		ldr     r12, [r1, #8]

	adds	r4, r4, r11
	adc		r5, r5, #0


	/* col 9 */
	umlal	r4, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #12]
		ldr		r14, [r2, #24]

	adds	r5, r5, r10
	adc		r6, r6, #0

	umlal	r4, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #16]
		ldr		r14, [r2, #20]

	adds	r5, r5, r11
	adc		r6, r6, #0

	umlal	r4, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #20]
		ldr		r14, [r2, #16]

	adds	r5, r5, r10
	adc		r6, r6, #0

	umlal	r4, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #24]
		ldr		r14, [r2, #12]

	adds	r5, r5, r11
	adc		r6, r6, #0

	umlal	r4, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #28]
		ldr		r14, [r2, #8]

	adds	r5, r5, r10
	adc		r6, r6, #0

	umlal	r4, r11, r12, r14
	
		mov		r10, #0
		ldr		r14, [r2, #12]

	adds	r5, r5, r11
	adc		r6, r6, #0

	
	/* col 10 */
	umlal	r5, r10, r12, r14

		mov		r11, #0
		ldr     r12, [r1, #24]
		ldr		r14, [r2, #16]

	adds	r6, r6, r10
	adc		r7, r7, #0

	umlal	r5, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #20]
		ldr		r14, [r2, #20]

	adds	r6, r6, r11
	adc		r7, r7, #0

	umlal	r5, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #16]
		ldr		r14, [r2, #24]

	adds	r6, r6, r10
	adc		r7, r7, #0

	umlal	r5, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #12]
		ldr		r14, [r2, #28]

	adds	r6, r6, r11
	adc		r7, r7, #0

	umlal	r5, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #16]

	adds	r6, r6, r10
	adc		r7, r7, #0


	/* col 11 */
	umlal	r6, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #20]
		ldr		r14, [r2, #24]

	adds	r7, r7, r11
	adc		r8, r8, #0

	umlal	r6, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #24]
		ldr		r14, [r2, #20]

	adds	r7, r7, r10
	adc		r8, r8, #0

	umlal	r6, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #28]
		ldr		r14, [r2, #16]

	adds	r7, r7, r11
	adc		r8, r8, #0

	umlal	r6, r10, r12, r14
	
		mov		r11, #0
		ldr		r14, [r2, #20]

			stmia	r0!, {r3-r6}

	adds	r7, r7, r10
	adc		r8, r8, #0

			mov		r3, r7

/*	mov		r7, #0
	mov		r8, #0 */
	
	/* col 12 */
	umlal	r3, r11, r12, r14

			mov		r4, r8
			mov		r5, #0
		mov		r10, #0
		ldr     r12, [r1, #24]
		ldr		r14, [r2, #24]

	adds	r4, r4, r11
	adc		r5, r5, #0

	umlal	r3, r10, r12, r14
	
			mov		r6, #0
		mov		r11, #0
		ldr     r12, [r1, #20]
		ldr		r14, [r2, #28]

	adds	r4, r4, r10
	adc		r5, r5, #0

	umlal	r3, r11, r12, r14
	
		mov		r10, #0
		ldr     r12, [r1, #24]

	adds	r4, r4, r11
	adc		r5, r5, #0

	/* col 13 */
	umlal	r4, r10, r12, r14
	
		mov		r11, #0
		ldr     r12, [r1, #28]
		ldr		r14, [r2, #24]

	adds	r5, r5, r10
	adc		r6, r6, #0

	umlal	r4, r11, r12, r14
	
/*		mov		r10, #0 */
		ldr		r14, [r2, #28]

	adds	r5, r5, r11
	adc		r6, r6, #0

	
	/* col 14 */
	umlal	r5, r6, r12, r14

/*	mov		r11, #0 */
	
	stmia	r0, {r3-r6}

	pop		{r14}
	pop		{r0-r12}
	mov		pc, lr
	
	.end
	